from __future__ import annotations

import importlib.util
import inspect
import json
import os
import sys
import tempfile
import uuid
import zipfile
from pathlib import Path
from time import sleep
from typing import Dict, TypedDict

import requests
from rich.console import Console

from gauge.cli.console import log_error, log_task

API_URL = os.environ.get("GAUGE_API_URL", "http://localhost:8000")


class DeployType(TypedDict):
    module: str | None
    reference: str
    function: str
    python_version: str
    dependencies: list[str]


DeployConfigType = Dict[str, DeployType]


class DeployHandler:
    def __init__(self, file_paths: list[str], deploy_name: str = "") -> None:
        self.file_paths = {Path(file_path) for file_path in file_paths}
        if not deploy_name:
            deploy_name = str(uuid.uuid4())
        self.deploy_name = deploy_name

    def validate_file_paths(self) -> None:
        errored = False
        for file_path in self.file_paths:
            if not file_path.exists():
                errored = True
                log_error(f"{file_path} does not exist")
            elif not file_path.is_file():
                errored = True
                log_error(f"{file_path} is not a file")
            elif file_path.suffix.lower() != ".py":
                errored = True
                log_error(f"{file_path} is not a python file")
        if errored:
            sys.exit(1)

    def bundle(self, temp_dir: str) -> Path:
        with log_task(start_message="Bundling...", end_message="Project bundled"):
            zip_filename = f"{self.deploy_name}.zip"
            zip_path = Path(temp_dir) / zip_filename

            with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
                for file_path in self.file_paths:
                    zipf.write(file_path)
        return zip_path

    def upload(self, zip_path: Path, deployments: DeployConfigType) -> None:
        print(deployments)
        gauge_client_id = os.environ.get("GAUGE_CLIENT_ID") or input(
            "Input your GAUGE_CLIENT_ID: "
        )
        with log_task(
            start_message="Uploading bundle...", end_message="Bundle uploaded"
        ):
            with open(zip_path, "rb") as zip_file:
                files = {"file": zip_file, "json_data": (None, json.dumps(deployments))}
                resp = requests.post(
                    API_URL + "/v0.1/deploy/",
                    headers={"GAUGE_CLIENT_ID": gauge_client_id},
                    files=files,
                )
            if resp.status_code != 200:
                print(resp.status_code, resp.content)
                log_error("Failed to trigger the deploy")
                sys.exit(1)

    def retrieve(self):
        with log_task(
            start_message="Checking deploy status...", end_message="Status updated"
        ):
            status = "pending"
            while status == "pending":
                resp = requests.get(API_URL)
                sleep(0.1)
                if resp.status_code == 200:
                    deployment_data = resp.json()
                    for deployment in deployment_data["deployments"]:
                        if deployment["name"] == self.deploy_name[:8]:
                            status = deployment["status"]
                            if status == "error":
                                log_error(deployment["debug"])
                            elif status == "deployed":
                                return deployment.get("deploy_url", "[deployment_url]")
                            break
                else:
                    sleep(3)

    def register_deployments(self) -> DeployConfigType:
        results = {}
        for file_path in self.file_paths:
            module_name = str(file_path).replace(os.path.sep, ".").replace(".py", "")
            spec = importlib.util.spec_from_file_location(file_path.stem, file_path)
            if spec and spec.loader:
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                for name, obj in inspect.getmembers(module):
                    if inspect.isfunction(obj) and hasattr(obj, "_gauge_register"):
                        try:
                            name, config = obj._gauge_register()  # type: ignore
                            if name in results:
                                raise ValueError(f"Duplicate endpoint {name}")
                            results[name] = {
                                "module": module_name,
                                "reference": f"{module_name}:{config['function']}",
                                **config,
                            }

                        except Exception as e:
                            print(f"Error calling _gauge_register on {name}: {e}")
                            break
            else:
                print(f"Couldn't load module from {file_path}")
        return results  # type: ignore

    def deploy(self):
        console = Console()
        console.print(
            f"Deploying [bold white]"
            f"[bold green] as [bold white]{self.deploy_name}[bold green]...",
        )
        self.validate_file_paths()
        deployments = self.register_deployments()
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_path = self.bundle(temp_dir)
            self.upload(zip_path, deployments)
        console.print(f"[bold white]{self.deploy_name[:8]} [bold green]deployed!")



from __future__ import annotations

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from gauge.cli.deploy import DeployHandler


@pytest.fixture
def deploy_handler() -> DeployHandler:
    return DeployHandler(["test_file.py"], "test_deploy")


def test_init() -> None:
    handler = DeployHandler(["file1.py", "file2.py"], "test_deploy")
    assert len(handler.file_paths) == 2
    assert handler.deploy_name == "test_deploy"

    handler_no_name = DeployHandler(["file1.py"])
    assert len(handler_no_name.file_paths) == 1
    assert len(handler_no_name.deploy_name) == 36  # UUID4 length


def test_validate_file_paths(deploy_handler: DeployHandler) -> None:
    with patch("pathlib.Path.exists", return_value=True), patch(
        "pathlib.Path.is_file", return_value=True
    ):
        deploy_handler.validate_file_paths()  # Should not raise any exception

    with pytest.raises(SystemExit):
        with patch("pathlib.Path.exists", return_value=False):
            deploy_handler.validate_file_paths()

    with pytest.raises(SystemExit):
        with patch("pathlib.Path.exists", return_value=True), patch(
            "pathlib.Path.is_file", return_value=False
        ):
            deploy_handler.validate_file_paths()

    with pytest.raises(SystemExit):
        deploy_handler.file_paths = {Path("test_file.txt")}
        deploy_handler.validate_file_paths()


def test_bundle(deploy_handler: DeployHandler) -> None:
    with patch("zipfile.ZipFile") as mock_zipfile:
        result = deploy_handler.bundle("/tmp")
        assert isinstance(result, Path)
        assert result.name == f"{deploy_handler.deploy_name}.zip"
        mock_zipfile.assert_called_once()


@patch.object(DeployHandler, "validate_file_paths")
@patch.object(DeployHandler, "register_deployments")
@patch.object(DeployHandler, "bundle")
@patch.object(DeployHandler, "upload")
def test_deploy(
    mock_upload: MagicMock,
    mock_bundle: MagicMock,
    mock_register: MagicMock,
    mock_validate: MagicMock,
    deploy_handler: DeployHandler,
) -> None:
    mock_register.return_value = {"test_endpoint": {"function": "test_func"}}
    mock_bundle.return_value = Path("test.zip")

    deploy_handler.deploy()

    mock_validate.assert_called_once()
    mock_register.assert_called_once()
    mock_bundle.assert_called_once()
    mock_upload.assert_called_once()



from __future__ import annotations

from gauge.sdk.main import endpoint


@endpoint(name="name1")
def hello_world():
    print("Hello World!")


@endpoint(name="name2", python_version="3.11")
def test():
    pass



from __future__ import annotations

import argparse

from gauge.cli.deploy import DeployHandler


def deploy(file_path_str: str) -> None:
    file_paths = [path.strip() for path in file_path_str.split(",")]
    DeployHandler(file_paths=file_paths, deploy_name="test").deploy()


def status() -> None:
    """Check the status of the application."""
    print("Checking the status of the application...")
    print("Status checked")


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="A CLI tool to deploy python lambdas with a single command"
    )

    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    parser_deploy = subparsers.add_parser("deploy", help="Deploy the application.")
    parser_deploy.add_argument(
        "file_paths", type=str, help="A comma-separated list of file paths to deploy."
    )

    subparsers.add_parser("status", help="Check the status of the application.")

    return parser


def main() -> None:
    parser = create_parser()
    args = parser.parse_args()
    if args.command == "deploy":
        deploy(args.file_paths)
    elif args.command == "status":
        status()
    else:
        print("Unknown command")
        parser.print_help()


if __name__ == "__main__":
    main()



from __future__ import annotations

from gauge import endpoint


def test_default_endpoint_decorator():
    @endpoint(name="name")
    def test_function():
        return "testy"

    assert test_function() == "testy"
    assert hasattr(test_function, "_gauge_register")
    assert test_function._gauge_register() == (
        "name",
        {
            "dependencies": [],
            "python_version": "3.12",
            "function": "test_function",
        },
    )


def test_custom_endpoint_decorator():
    @endpoint(name="name2", dependencies=["fastapi"], python_version="3.13")
    def test_function():
        return "test2"

    assert test_function() == "test2"
    assert hasattr(test_function, "_gauge_register")
    assert test_function._gauge_register() == (
        "name2",
        {
            "dependencies": ["fastapi"],
            "python_version": "3.13",
            "function": "test_function",
        },
    )



from __future__ import annotations

from typing import Any, Callable


def endpoint(
    name: str, python_version: str = "3.12", dependencies: list[str] = []
) -> Callable[..., Any]:
    def endpoint_decorator(
        function: Callable[..., Any],
    ) -> Callable[..., Any]:
        def _gauge_register() -> tuple[str, dict[str, str | list[str]]]:
            return name, {
                "function": function.__name__,
                "python_version": python_version,
                "dependencies": dependencies,
            }

        function._gauge_register = _gauge_register  # pyright: ignore[reportFunctionMemberAccess]
        return function

    return endpoint_decorator
